'''
show_route.py

'''
import re
from genie.metaparser import MetaParser
from genie.metaparser.util.schemaengine import Schema, \
                                         Any, \
                                         Optional

# ====================================================
#  distributor class for show ip route
# ====================================================
class ShowIpRouteDistributor(MetaParser):
    """distributor class for show ip route"""
    cli_command = ['show ip route vrf {vrf}',
                   'show ip route vrf {vrf} {route}',
                   'show ip route vrf {vrf} {protocol}',
                   'show ip route',
                   'show ip route {route}',
                   'show ip route {protocol}']

    protocol_set = {'ospf', 'odr', 'isis', 'eigrp', 'static', 'mobile',
                    'rip', 'lisp', 'nhrp', 'local', 'connected', 'bgp'}

    def cli(self, vrf='', route='', protocol='', output=None):
        if not vrf:
            vrf = 'default'
        if output is None:
            if vrf != 'default':
                if route:
                    cmd = self.cli_command[1].format(vrf=vrf, route=route)
                elif protocol:
                    cmd = self.cli_command[2].format(vrf=vrf, protocol=protocol)
                else:
                    cmd = self.cli_command[0].format(vrf=vrf)
            else:
                if route:
                    cmd = self.cli_command[4].format(route=route)
                elif protocol:
                    cmd = self.cli_command[5].format(protocol=protocol)
                else:
                    cmd = self.cli_command[3]

            out = self.device.execute(cmd)
        else:
            out = output

        if (route or protocol) in self.protocol_set or (not route and not protocol):
            parser = ShowIpRoute(self.device)
            self.schema = parser.schema
            return parser.parse(vrf=vrf, protocol=protocol, output=out)

        else:
            parser = ShowIpRouteWord(self.device)
            self.schema=parser.schema
            return parser.parse(vrf=vrf, route=route, output=out)

# ====================================================
#  distributor class for show ipv6 route
# ====================================================
class ShowIpv6RouteDistributor(MetaParser):
    """distributor class for show ipv6 route"""
    cli_command = ['show ipv6 route vrf {vrf}',
                   'show ipv6 route vrf {vrf} {route}',
                   'show ipv6 route vrf {vrf} {protocol}',
                   'show ipv6 route',
                   'show ipv6 route {route}',
                   'show ipv6 route {protocol}']

    protocol_set = {'ospf', 'odr', 'isis', 'eigrp', 'static', 'mobile',
                    'rip', 'lisp', 'nhrp', 'local', 'connected', 'bgp'}

    def cli(self, vrf='', route='', protocol='', output=None):
        if not vrf:
            vrf = 'default'
        if output is None:
            if vrf != 'default':
                if route:
                    cmd = self.cli_command[1].format(vrf=vrf, route=route)
                elif protocol:
                    cmd = self.cli_command[2].format(vrf=vrf, protocol=protocol)
                else:
                    cmd = self.cli_command[0].format(vrf=vrf)
            else:
                if route:
                    cmd = self.cli_command[4].format(route=route)
                elif protocol:
                    cmd = self.cli_command[5].format(protocol=protocol)
                else:
                    cmd = self.cli_command[3]

            out = self.device.execute(cmd)
        else:
            out = output

        if (route or protocol) in self.protocol_set or (not route and not protocol):
            parser = ShowIpv6Route(self.device)
            self.schema = parser.schema
            return parser.parse(vrf=vrf, protocol=protocol, output=out)

        else:
            parser = ShowIpv6RouteWord(self.device)
            self.schema=parser.schema
            return parser.parse(vrf=vrf, route=route, output=out)

# ====================================================
#  schema for show ip route
# ====================================================
class ShowIpRouteSchema(MetaParser):
    """Schema for show ip route"""
    schema = {
        'vrf': {
            Any(): {
                Optional('address_family'): {
                    Any(): {
                        Optional('routes'): {
                            Any(): {
                                Optional('route'): str,
                                Optional('active'): bool,
                                Optional('route_preference'): int,
                                Optional('metric'): int,
                                Optional('source_protocol'): str,
                                Optional('source_protocol_codes'): str,
                                Optional('mask'): str,
                                Optional('known_via'): str,
                                Optional('distance'): int,
                                Optional('type'): str,
                                Optional('net'): str,
                                Optional('redist_via'): str,
                                Optional('redist_via_tag'): str,
                                Optional('update'): {
                                    'from': str,
                                    'interface': str,
                                    'age': str
                                },
                                Optional('next_hop'): {
                                    Optional('outgoing_interface'): {
                                        Any(): {  # interface  if there is no next_hop
                                            Optional('outgoing_interface'): str,
                                            Optional('updated'): str,
                                        },
                                    },
                                    Optional('next_hop_list'): {
                                        Any(): {  # index
                                            Optional('index'): int,
                                            Optional('next_hop'): str,
                                            Optional('outgoing_interface'): str,
                                            Optional('updated'): str,
                                            Optional('age'): str,
                                            Optional('from'): str,
                                            Optional('metric'): str,
                                            Optional('share_count'): str,
                                            Optional('loading'): str,
                                            Optional('hops'): str,
                                            Optional('minimum_mtu'): str,
                                            Optional('reliability'): str,
                                            Optional('minimum_bandwidth'): str,
                                            Optional('total_delay'): str,
                                            Optional('vrf'): str
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    }


# ====================================================
#  parser for show ip route
# ====================================================
class ShowIpRoute(ShowIpRouteSchema):
    """Parser for :
        show ip route
        show ip route vrf <vrf>"""
    # not using name 'cli_command' because dont want find_parsers() to discover them
    command = ['show ip route vrf {vrf}', 'show ip route vrf {vrf} {protocol}',
                   'show ip route', 'show ip route {protocol}']
    exclude = ['updated']
    IP_VER='ipv4'

    def cli(self, vrf="", protocol='', output=None):
        if not vrf:
            vrf = 'default'
        if output is None:
            if vrf != 'default':
                if protocol:
                    cmd = self.command[1].format(vrf=vrf, protocol=protocol)
                else:
                    cmd = self.command[0].format(vrf=vrf)
            else:
                if protocol:
                    cmd = self.command[3].format(protocol=protocol)
                else:
                    cmd = self.command[2]
            out = self.device.execute(cmd)
        else:
            out = output

        af = self.IP_VER
        route = ""
        source_protocol_dict = {}
        source_protocol_dict['ospf'] = ['O','IA','N1','N2','E1','E2']
        source_protocol_dict['odr'] = ['o']
        source_protocol_dict['isis'] = ['i','su','L1','L2','ia', 'I1', 'I2']
        source_protocol_dict['eigrp'] = ['D','EX']
        source_protocol_dict['static'] = ['S']
        source_protocol_dict['mobile'] = ['M']
        source_protocol_dict['rip'] = ['R']
        source_protocol_dict['lisp'] = ['I', 'Ir','Ia','Id']
        source_protocol_dict['nhrp'] = ['H']
        source_protocol_dict['local'] = ['L']
        source_protocol_dict['connected'] = ['C']
        source_protocol_dict['local_connected'] = ['LC']
        source_protocol_dict['bgp'] = ['B']

        result_dict = {}

        # initial regexp pattern
        p100 = re.compile(r'^Routing +entry +for +'
                        '(?P<entry>(?P<ip>[\w\:\.]+)\/(?P<mask>\d+))'
                        '(, +(?P<net>[\w\s]+))?$')
        p200 = re.compile(r'^Known +via +\"(?P<known_via>[\w\s]+)\", +'
                        'distance +(?P<distance>\d+), +'
                        'metric +(?P<metric>\d+)'
                        '(, +type +(?P<type>[\w\-\s]+)(?P<connected>, connected)?)?$')
        p300 = re.compile(r'^Redistributing +via +(?P<redist_via>\w+) *'
                        '(?P<redist_via_tag>\d+)?$')
        p400 = re.compile(r'^Last +update +from +(?P<from>[\w\.]+) +'
                        'on +(?P<interface>[\w\.\/\-]+), +'
                        '(?P<age>[\w\.\:]+) +ago$')
        p500 = re.compile(r'^\*? *(?P<nexthop>[\w\.]+)(, +'
                        'from +(?P<from>[\w\.]+), +'
                        '(?P<age>[\w\.\:]+) +ago, +'
                        'via +(?P<interface>[\w\.\/\-]+))?$')
        p600 = re.compile(r'^Route +metric +is +(?P<metric>\d+), +'
                        'traffic +share +count +is +(?P<share_count>\d+)$')

        p600 = re.compile(r'^Route +metric +is +(?P<metric>\d+), +'
                          'traffic +share +count +is +(?P<share_count>\d+)$')
        p700 = re.compile(r'^Total +delay +is +(?P<total_delay>\d+) +microseconds, '
                          '+minimum +bandwidth +is +(?P<minimum_bandwidth>\d+) +Kbit$')
        p800 = re.compile(r'^Reliability +(?P<reliability>[\d\/]+), +minimum +MTU +(?P<minimum_mtu>\d+) +bytes$')
        p900 = re.compile(r'^Loading +(?P<loading>[\d\/]+), Hops +(?P<hops>\d+)$')

        # initial variables
        ret_dict = {}
        index = 0


        for line in out.splitlines():
            if line:
                line = line.strip()
            else:
                continue
            next_hop = interface = updated = metrics = route_preference = ""
            # Routing Table: VRF1
            p1 = re.compile(r'^\s*Routing Table: +(?P<vrf>[\w]+)$')
            m = p1.match(line)
            if m:
                vrf = m.groupdict()['vrf']
                continue

            # 10.1.0.0/32 is subnetted, 1 subnets
            # 10.0.0.0/8 is variably subnetted, 5 subnets, 2 masks
            p2 = re.compile(r'^\s*(?P<subnetted_ip>[\d\/\.]+)'
                            ' +is +(variably )?subnetted, +(?P<number_of_subnets>[\d]+) +subnets(, +(?P<number_of_masks>[\d]+) +masks)?$')
            m = p2.match(line)
            if m:
                # if you see the issue by "show ip route", it means that active is True.
                # it means all routes in the output should be active=True
                active = True
                netmask = number_of_masks= ""
                number_of_subnets = m.groupdict()['number_of_subnets']
                if m.groupdict()['number_of_masks']:
                    number_of_masks = m.groupdict()['number_of_masks']

                if m.groupdict()['subnetted_ip']:
                    subnetted_ip = m.groupdict()['subnetted_ip']
                    if '/' in subnetted_ip:
                        netmask = subnetted_ip.split('/')[1]
                continue

            # C        10.4.1.1 is directly connected, Loopback0
            # S        10.16.2.2 [1/0] via 10.186.2.2, GigabitEthernet0/1
            # S*       10.16.2.2 [1/0] via 10.186.2.2, GigabitEthernet0/1
            # O        10.2.3.0/24 [110/2] via 10.186.2.2, 06:46:59, GigabitEthernet0/1
            # i L1     10.151.22.22 [115/20] via 10.186.2.2, 06:47:04, GigabitEthernet0/1
            # D        192.168.205.1
            # S*       0.0.0.0/0 [1/0] via 10.50.15.1
            p3 = re.compile(
                r'^\s*(?P<code>[\w\*]+) +(?P<code1>[\w]+)? +(?P<network>[0-9\.\:\/]+)?( '
                r'+is +directly +connected,)? *\[?(?P<route_preference>[\d\/]+)?\]?( *('
                r'via +)?(?P<next_hop>[\d\.]+))?,?( +(?P<date>[0-9][\w\:]+))?,?( +(?P<interface>[\S]+))?$')

            m = p3.match(line)
            if m:
                active = True
                if m.groupdict()['code']:
                    source_protocol_codes = m.groupdict()['code'].strip()
                    for key,val in source_protocol_dict.items():
                        source_protocol_replaced = source_protocol_codes.split('*')[0]
                        if source_protocol_replaced in val:
                            source_protocol = key

                if m.groupdict()['code1']:
                    source_protocol_codes = '{} {}'.format(source_protocol_codes, m.groupdict()['code1'])

                if m.groupdict()['network']:
                    network = m.groupdict()['network']
                    if '/' in network:
                        route = network
                    else:
                        route = '{}/{}'.format(network,netmask)

                if not m.groupdict()['network']:
                    route = route
                if m.groupdict()['route_preference']:
                    routepreference = m.groupdict()['route_preference']
                    if '/' in routepreference:
                        route_preference = int(routepreference.split('/')[0])
                        metrics = routepreference.split('/')[1]


                if m.groupdict()['next_hop']:
                    next_hop = m.groupdict()['next_hop']
                    index = 1
                else:
                    index = 0
                if m.groupdict()['interface']:
                    interface = m.groupdict()['interface']

                if m.groupdict()['date']:
                    updated = m.groupdict()['date']

                if vrf:
                    if 'vrf' not in result_dict:
                        result_dict['vrf'] = {}

                    if vrf not in result_dict['vrf']:
                        result_dict['vrf'][vrf] = {}

                    if 'address_family' not in result_dict['vrf'][vrf]:
                        result_dict['vrf'][vrf]['address_family'] = {}

                    if af and af not in result_dict['vrf'][vrf]['address_family']:
                        result_dict['vrf'][vrf]['address_family'][af] = {}

                    if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                    if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['route'] = route

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['active'] = active

                    if metrics:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['metric'] = int(metrics)
                    if route_preference:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['route_preference'] = route_preference
                    if source_protocol_codes:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['source_protocol_codes'] = source_protocol_codes
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['source_protocol'] = source_protocol

                    if 'next_hop' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] = {}

                    if not next_hop and interface:

                        if 'outgoing_interface' not in result_dict['vrf'][vrf]['address_family'][af] \
                                ['routes'][route]['next_hop']:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]\
                                ['next_hop']['outgoing_interface'] = {}

                        if m.groupdict()['interface'] and interface not in \
                                result_dict['vrf'][vrf]['address_family'][af]['routes'][route]\
                                ['next_hop']['outgoing_interface']:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]\
                                ['next_hop']['outgoing_interface'][interface] = {}

                        if interface:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                                ['next_hop']['outgoing_interface'][interface]['outgoing_interface'] = interface

                    elif next_hop:
                        if 'next_hop_list' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop']:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'][
                                'next_hop_list'] = {}


                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index] = {}

                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index]['index'] = index
                        if next_hop:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                                ['next_hop_list'][index]['next_hop'] = next_hop

                        if updated:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                                ['next_hop_list'][index]['updated'] = updated

                        if interface:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                                ['next_hop_list'][index]['outgoing_interface'] = interface

                continue

            #    [110/2] via 10.1.2.2, 06:46:59, GigabitEthernet0/0
            p4 = re.compile(r'^\s*\[(?P<route_preference>[\d\/]+)\]'
                            ' +via +(?P<next_hop>[\d\.]+)?,?( +(?P<date>[0-9][\w\:]+),)?( +(?P<interface>[\S]+))?$')
            m = p4.match(line)
            if m:

                routepreference = m.groupdict()['route_preference']
                if routepreference and '/' in routepreference:
                    route_preference = int(routepreference.split('/')[0])
                    metrics = routepreference.split('/')[1]

                next_hop = m.groupdict()['next_hop']
                index +=1
                if m.groupdict()['interface']:
                    interface = m.groupdict()['interface']

                if m.groupdict()['date']:
                    updated = m.groupdict()['date']

                if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['route'] = route

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                    ['active'] = active

                if metrics:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['metric'] = int(metrics)
                if route_preference:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['route_preference'] = route_preference
                if source_protocol_codes:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['source_protocol_codes'] = source_protocol_codes
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['source_protocol'] = source_protocol

                if 'next_hop' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] = {}

                if not next_hop:
                    if 'outgoing_interface' not in result_dict['vrf'][vrf]['address_family'][af] \
                            ['routes'][route]['next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'] = {}

                    if m.groupdict()['interface'] and interface not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'][interface] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['next_hop']['outgoing_interface'][interface]['outgoing_interface'] = interface

                else:
                    if 'next_hop_list' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'][
                            'next_hop_list'] = {}

                    if index not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                        ['next_hop_list'][index]['index'] = index

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                        ['next_hop_list'][index]['next_hop'] = next_hop

                    if updated:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index]['updated'] = updated

                    if interface:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index]['outgoing_interface'] = interface

                continue

            #       is directly connected, GigabitEthernet0/2
            p5 = re.compile(r'^\s*is +directly +connected,( +\[(?P<route_preference>[\d\/]+)\]'
                            ' +via +(?P<next_hop>[\d\.]+)?,)?( +(?P<date>[0-9][\w\:]+),)?( +(?P<interface>[\S]+))?$')
            m = p5.match(line)
            if m:

                if m.groupdict()['route_preference']:
                    routepreference = m.groupdict()['route_preference']
                    if '/' in routepreference:
                        route_preference = int(routepreference.split('/')[0])
                        metrics = routepreference.split('/')[1]

                index += 1
                if m.groupdict()['next_hop']:
                    next_hop = m.groupdict()['next_hop']
                if m.groupdict()['interface']:
                    interface = m.groupdict()['interface']

                if m.groupdict()['date']:
                    updated = m.groupdict()['date']

                if vrf:
                    if 'vrf' not in result_dict:
                        result_dict['vrf'] = {}

                    if vrf not in result_dict['vrf']:
                        result_dict['vrf'][vrf] = {}

                    if 'address_family' not in result_dict['vrf'][vrf]:
                        result_dict['vrf'][vrf]['address_family'] = {}

                    if af and af not in result_dict['vrf'][vrf]['address_family']:
                        result_dict['vrf'][vrf]['address_family'][af] = {}

                    if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                    if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['route'] = route


                    if metrics:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['metric'] = int(metrics)
                    if route_preference:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['route_preference'] = route_preference

                    if 'next_hop' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] = {}

                    if not next_hop:

                        if 'outgoing_interface' not in result_dict['vrf'][vrf]['address_family'][af] \
                                ['routes'][route]['next_hop']:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                                ['next_hop']['outgoing_interface'] = {}

                        if m.groupdict()['interface'] and interface not in \
                                result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                                        ['next_hop']['outgoing_interface']:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                                ['next_hop']['outgoing_interface'][interface] = {}

                        if interface:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                                ['next_hop']['outgoing_interface'][interface]['outgoing_interface'] = interface

                    else:
                        if 'next_hop_list' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop']:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'][
                                'next_hop_list'] = {}

                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index] = {}

                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index]['index'] = index

                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index]['next_hop'] = next_hop

                        if updated:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                                ['next_hop_list'][index]['updated'] = updated

                        if interface:
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                                ['next_hop_list'][index]['outgoing_interface'] = interface

                continue
            #      via 2001:DB8:1:1::2
            #      via 10.4.1.1%default, indirectly connected
            #      via 2001:DB8:4:6::6
            #      via 2001:DB8:20:4:6::6%VRF2
            p6 = re.compile(r'^\s*via +(?P<next_hop>[a-zA-Z0-9./:]+)%?(?P<vrf>[a-zA-Z0-9]+)?,?[\w\s]*$')
            m = p6.match(line)
            if m:
                vrf_val = ''
                next_hop = m.groupdict()['next_hop']
                if m.groupdict()['vrf']:
                    vrf_val = m.groupdict()['vrf']
                index += 1
                if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                    'route'] = route

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                    ['active'] = active

                if metrics:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['metric'] = int(metrics)
                if route_preference:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['route_preference'] = route_preference
                if source_protocol_codes:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['source_protocol_codes'] = source_protocol_codes
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['source_protocol'] = source_protocol

                if 'next_hop' not in \
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                        'next_hop'] = {}

                if 'next_hop_list' not in \
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                            'next_hop']:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                        'next_hop'][
                        'next_hop_list'] = {}

                if index not in \
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                            'next_hop'] \
                                ['next_hop_list']:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                        'next_hop'] \
                        ['next_hop_list'][index] = {}

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                    'next_hop'] \
                    ['next_hop_list'][index]['index'] = index

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                    'next_hop'] \
                    ['next_hop_list'][index]['next_hop'] = next_hop

                if vrf_val:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                        'next_hop'] \
                        ['next_hop_list'][index]['vrf'] = vrf_val

                continue
            # Routing entry for 10.151.0.0/24, 1 known subnets
            # Routing entry for 0.0.0.0/0, supernet
            # Routing entry for 192.168.154.0/24
            m = p100.match(line)
            if m:
                group = m.groupdict()
                entry_dict = result_dict.setdefault('vrf', {}).setdefault(vrf, {}).setdefault('address_family',
                                                                                              {}).setdefault(af, {})
                route_dict = entry_dict.setdefault('routes', {}).setdefault(route, {})
                route_dict.update({'route': group['ip']})
                route_dict.update({'mask': group['mask']})
                route_dict.update({'active': True})
                continue

            # Known via "eigrp 1", distance 130, metric 10880, type internal
            # Known via "rip", distance 120, metric 2
            m = p200.match(line)
            if m:
                group = m.groupdict()
                route_dict.update({'distance': int(group['distance'])})
                route_dict.update({'metric': int(group['metric'])})
                if group['type']:
                    route_dict.update({'type': group['type']})
                continue

            # Redistributing via rip
            # Redistributing via eigrp 1
            m = p300.match(line)
            if m:
                group = m.groupdict()
                route_dict.update({k: v for k, v in group.items() if v})
                continue

            # Last update from 192.168.151.2 on Vlan101, 2w3d ago
            # Last update from 192.168.246.2 on Vlan103, 00:00:12 ago
            m = p400.match(line)
            if m:
                group = m.groupdict()
                update_dict = route_dict.setdefault('update', {})
                update_dict.update({k: v for k, v in group.items() if v})
                continue

            # * 192.168.151.2, from 192.168.151.2, 2w3d ago, via Vlan101
            # * 10.69.1.2
            m = p500.match(line)
            if m:
                group = m.groupdict()
                index += 1
                path_dict = route_dict.setdefault('next_hop',{}).setdefault('next_hop_list', {}).setdefault(index, {})
                path_dict.update({'index': index})
                path_dict.update({'next_hop': group['nexthop']})
                path_dict.update({'age': group['age']})
                path_dict.update({'from': group['from']})
                path_dict.update({'outgoing_interface': group['interface']})
                continue

            # Route metric is 10880, traffic share count is 1
            m = p600.match(line)
            if m:
                group = m.groupdict()
                path_dict.update({k: v for k, v in group.items() if v})

            # Total delay is 20 microseconds, minimum bandwidth is 1000000 Kbit
            m = p700.match(line)
            if m:
                group = m.groupdict()
                path_dict.update({k: v for k, v in group.items() if v})
                continue

            # Reliability 255/255, minimum MTU 1500 bytes
            m = p800.match(line)
            if m:
                group = m.groupdict()
                path_dict.update({k: v for k, v in group.items() if v})
                continue

            # Loading 1/255, Hops 1
            m = p900.match(line)
            if m:
                group = m.groupdict()
                path_dict.update({k: v for k, v in group.items() if v})
                continue
        return result_dict

class ShowIpv6Route(ShowIpRoute):
    """Parser for:
        show ipv6 route
        show ipv6 route vrf <vrf>"""
    command = ['show ipv6 route vrf {vrf}', 'show ipv6 route vrf {vrf} {protocol}',
               'show ipv6 route', 'show ipv6 route {protocol}']
    exclude = ['uptime']

    IP_VER = 'ipv6'
    def cli(self, vrf='', protocol='', output=None):
        if not vrf:
            vrf = 'default'
        if output is None:
            if vrf != 'default':
                if protocol:
                    cmd = self.command[1].format(vrf=vrf, protocol=protocol)
                else:
                    cmd = self.command[0].format(vrf=vrf)
            else:
                if protocol:
                    cmd = self.command[3].format(protocol=protocol)
                else:
                    cmd = self.command[2]
            out = self.device.execute(cmd)
        else:
            out = output
        return super().cli(vrf=vrf, protocol=protocol, output=out)

# ====================================================
#  schema for show ipv6 route updated
# ====================================================
class ShowIpv6RouteUpdatedSchema(MetaParser):
    """Schema for show ipv6 route updated"""
    schema = {
        'ipv6_unicast_routing_enabled': bool,
        'vrf': {
            Any(): {
                Optional('address_family'): {
                    Any(): {
                        Optional('routes'): {
                            Any(): {
                                Optional('route'): str,
                                Optional('active'): bool,
                                Optional('route_preference'): int,
                                Optional('metric'): int,
                                Optional('source_protocol'): str,
                                Optional('source_protocol_codes'): str,
                                Optional('next_hop'): {
                                    Optional('outgoing_interface'): {
                                        Any(): {  # interface  if there is no next_hop
                                            Optional('outgoing_interface'): str,
                                            Optional('updated'): str,
                                        },
                                    },
                                    Optional('next_hop_list'): {
                                        Any(): {  # index
                                            Optional('index'): int,
                                            Optional('next_hop'): str,
                                            Optional('outgoing_interface'): str,
                                            Optional('updated'): str,
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    }


# ====================================================
#  parser for show ipv6 route updated
# ====================================================
class ShowIpv6RouteUpdated(ShowIpv6RouteUpdatedSchema):
    """Parser for :
       show ipv6 route updated
       show ipv6 route vrf <vrf> updated"""
    exclude = ['updated']

    cli_command = ['show ipv6 route vrf {vrf} updated', 'show ipv6 route updated']

    def cli(self, vrf="", output=None):
        if output is None:
            if vrf:
                cmd = self.cli_command[0].format(vrf=vrf)
            else:
                cmd = self.cli_command[1]
                vrf = 'default'
            out = self.device.execute(cmd)
        else:
            out = output

        af = 'ipv6'
        route = ""
        next_hop = interface = metrics = route_preference = ""
        source_protocol_dict = {}
        source_protocol_dict['ospf'] = ['O','OI','ON1','ON2','OE1','OE2']
        source_protocol_dict['isis'] = ['IS','I1','I2','IA']
        source_protocol_dict['eigrp'] = ['D','EX']
        source_protocol_dict['static'] = ['S']
        source_protocol_dict['mobile'] = ['M']
        source_protocol_dict['rip'] = ['R']
        source_protocol_dict['lisp'] = ['Ir','Ia','Id']
        source_protocol_dict['nhrp'] = ['H']
        source_protocol_dict['local'] = ['L']
        source_protocol_dict['connected'] = ['C']
        source_protocol_dict['bgp'] = ['B']
        source_protocol_dict['static route'] = ['U']
        source_protocol_dict['home agent'] = ['HA']
        source_protocol_dict['mobile router'] = ['MR']
        source_protocol_dict['nemo'] = ['NM']
        source_protocol_dict['nd'] = ['ND','NDp']
        source_protocol_dict['destination'] = ['DCE']
        source_protocol_dict['redirect'] = ['NDr']

        result_dict = {}
        for line in out.splitlines():
            if line:
                line = line.rstrip()
            else:
                continue

            # IPv6 Routing Table - default - 23 entries
            # IPv6 Routing Table - VRF1 - 104 entries
            p1 = re.compile(r'^\s*IPv6 +Routing +Table +\- +(?P<vrf>[\w]+) +\- +(?P<entries>[\d]+) +entries$')
            m = p1.match(line)
            if m:
                vrf = m.groupdict()['vrf']
                continue

            # LC  2001:1:1:1::1/128 [0/0]
            p2 = re.compile(r'^\s*(?P<code>[\w]+) +(?P<route>[\w\/\:]+)?'
                            ' +\[(?P<route_preference>[\d\/]+)\]$')
            m = p2.match(line)
            if m:
                active = True
                next_hop = interface = ""
                if m.groupdict()['code']:
                    source_protocol_codes = m.groupdict()['code'].strip()

                    for key, val in source_protocol_dict.items():
                        if source_protocol_codes in val:
                            source_protocol = key
                            break
                        else:
                            if 'L' in source_protocol_codes:
                                source_protocol = 'local'
                            # else:
                            #    source_protocol = source_protocol_codes

                if m.groupdict()['route']:
                    route = m.groupdict()['route']

                if m.groupdict()['route_preference']:
                    routepreference = m.groupdict()['route_preference']
                    if '/' in routepreference:
                        route_preference = routepreference.split('/')[0]
                        metrics = routepreference.split('/')[1]
                index = 1

                if vrf:
                    if 'vrf' not in result_dict:
                        result_dict['vrf'] = {}

                    if vrf not in result_dict['vrf']:
                        result_dict['vrf'][vrf] = {}

                    if 'address_family' not in result_dict['vrf'][vrf]:
                        result_dict['vrf'][vrf]['address_family'] = {}

                    if af and af not in result_dict['vrf'][vrf]['address_family']:
                        result_dict['vrf'][vrf]['address_family'][af] = {}

                    if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                    if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['route'] = route

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                        ['active'] = active

                    if metrics:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['metric'] = int(metrics)
                    if route_preference:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['route_preference'] = int(route_preference)
                    if source_protocol_codes:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['source_protocol_codes'] = source_protocol_codes
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['source_protocol'] = source_protocol
                continue

            #   via Loopback0, receive
            #   via 2001:10:1:2::2, GigabitEthernet0/0
            #   via GigabitEthernet0/2, directly connected
            #   via 192.168.51.1%default, indirectly connected
            p3 = re.compile(r'^\s*via( +(?P<next_hop>[0-9][\w\:\.\%]+)?,)?'
                            '( +(?P<interface>[\w\.\/\-\_]+))?,?( +receive)?( +directly connected)?( +indirectly connected)?$')
            m = p3.match(line)
            if m:
                if m.groupdict()['next_hop']:
                    if '%' in  m.groupdict()['next_hop']:
                        next_hop = m.groupdict()['next_hop'].split('%')[0]
                    else:
                        next_hop = m.groupdict()['next_hop']

                if m.groupdict()['interface']:
                    interface = m.groupdict()['interface']

                if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['route'] = route


                if 'next_hop' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] = {}

                if not next_hop:
                    if 'outgoing_interface' not in result_dict['vrf'][vrf]['address_family'][af] \
                            ['routes'][route]['next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'] = {}

                    if m.groupdict()['interface'] and interface not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'][interface] = {}

                    if interface:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'][interface]['outgoing_interface'] = interface

                else:
                    if 'next_hop_list' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'][
                            'next_hop_list'] = {}

                    if index not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                        ['next_hop_list'][index]['index'] = index

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                        ['next_hop_list'][index]['next_hop'] = next_hop

                    if interface:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index]['outgoing_interface'] = interface

                continue

            #   via FE80::211:1FF:FE00:1, GigabitEthernet0/0/2.100
            p4 = re.compile(r'^\s*via +(?P<next_hop>[\w\:\.\%]+),'
                            ' +(?P<interface>[\S]+)$')
            m = p4.match(line)
            if m:
                if m.groupdict()['next_hop']:
                    if '%' in m.groupdict()['next_hop']:
                        next_hop = m.groupdict()['next_hop'].split('%')[0]
                    else:
                        next_hop = m.groupdict()['next_hop']

                if m.groupdict()['interface']:
                    interface = m.groupdict()['interface']

                if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['route'] = route

                if 'next_hop' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] = {}

                if not next_hop:
                    if 'outgoing_interface' not in result_dict['vrf'][vrf]['address_family'][af] \
                            ['routes'][route]['next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'] = {}

                    if m.groupdict()['interface'] and interface not in \
                            result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                                    ['next_hop']['outgoing_interface']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'][interface] = {}

                    if interface:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'][interface]['outgoing_interface'] = interface

                else:
                    if 'next_hop_list' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route][
                        'next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'][
                            'next_hop_list'] = {}

                    if index not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                        ['next_hop_list'][index]['index'] = index

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                        ['next_hop_list'][index]['next_hop'] = next_hop

                    if interface:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index]['outgoing_interface'] = interface

                continue
            #      Last updated 14:15:23 06 December 2017
            p5 = re.compile(r'^\s*Last +updated +(?P<last_updated>[\S\s]+)$')
            m = p5.match(line)
            if m:

                last_updated = m.groupdict()['last_updated']

                if 'routes' not in result_dict['vrf'][vrf]['address_family'][af]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'] = {}
                if route not in result_dict['vrf'][vrf]['address_family'][af]['routes']:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route] = {}

                result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['route'] = route


                if 'next_hop' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]:
                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] = {}


                if not next_hop:
                    if 'outgoing_interface' not in result_dict['vrf'][vrf]['address_family'][af] \
                            ['routes'][route]['next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'] = {}

                    if interface and interface not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route] \
                            ['next_hop']['outgoing_interface'][interface] = {}

                    if interface:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]\
                        ['next_hop']['outgoing_interface'][interface]['updated'] = last_updated

                else:

                    if 'next_hop_list' not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'][
                            'next_hop_list'] = {}

                    if index not in result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list']:
                        result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop'] \
                            ['next_hop_list'][index] = {}

                    result_dict['vrf'][vrf]['address_family'][af]['routes'][route]['next_hop']\
                        ['next_hop_list'][index]['updated'] = last_updated

                index += 1

                continue

        if len(result_dict):
            result_dict['ipv6_unicast_routing_enabled'] = True
        return result_dict


# ====================================================
#  schema for show ip route <WORD>
# ====================================================
class ShowIpRouteWordSchema(MetaParser):
    """Schema for show ip route <WORD>"""
    schema = {
        'entry': {
            Any(): {
                'ip': str,
                'mask': str,
                'known_via': str,
                'distance': str,
                'metric': str,
                Optional('type'): str,
                Optional('net'): str,
                Optional('redist_via'): str,
                Optional('redist_via_tag'): str,
                Optional('sr_incoming_label'): str,
                Optional('update'): {
                    'from': str,
                    'interface': str,
                    'age': str
                },
                'paths': {
                    Any(): {
                        Optional('nexthop'): str,
                        Optional('from'): str,
                        Optional('age'): str,
                        Optional('interface'): str,
                        Optional('metric'): str,
                        Optional('share_count'): str,
                        Optional('mpls_label'): str,
                        Optional('mpls_flags'): str,
                        'prefer_non_rib_labels': bool,
                        'merge_labels': bool,
                        Optional('repair_path'): {
                            'repair_path': str,
                            'via': str
                        }
                    }
                }
            }
        },
        'total_prefixes': int,
    }


# ====================================================
#  parser for show ip route <WORD>
# ====================================================
class ShowIpRouteWord(ShowIpRouteWordSchema):
    """Parser for :
       show ip route <Hostname or A.B.C.D>
       show ip route vrf <vrf> <Hostname or A.B.C.D>"""
    command = ['show ip route vrf {vrf}', 'show ip route vrf {vrf} {route}',
                   'show ip route', 'show ip route {route}']
    IP_VER = 'ip'

    def cli(self, route='', vrf='', output=None):
        if not vrf:
            vrf = 'default'
        if output is None:
            if vrf != 'default':
                if route:
                    cmd = self.command[1].format(vrf=vrf, route=route)
                else:
                    cmd = self.command[0].format(vrf=vrf)
            else:
                if route:
                    cmd = self.command[3].format(route=route)
                else:
                    cmd = self.command[2]
            out = self.device.execute(cmd)
        else:
            out = output

        # initial regexp pattern
        p1 = re.compile(r'^Routing +entry +for +'
                         '(?P<entry>(?P<ip>[\w\:\.]+)\/(?P<mask>\d+))'
                         '(, +(?P<net>[\w\s]+))?$')

        # Known via "connected", distance 0, metric 0 (connected)
        # Known via "eigrp 1", distance 130, metric 10880, type internal
        p2 = re.compile(r'^Known +via +\"(?P<known_via>[\w\s]+)\", +distance +(?P<distance>\d+), +metric +(?P<metric>\d+),? *(?:type (?P<type>[\w\- ]+))?,? *.*$')
        p3 = re.compile(r'^Redistributing +via +(?P<redist_via>\w+) *'
                         '(?P<redist_via_tag>\d+)?$')
        p4 = re.compile(r'^Last +update +from +(?P<from>[\w\.]+) +'
                         'on +(?P<interface>[\w\.\/\-]+), +'
                         '(?P<age>[\w\.\:]+) +ago$')

        # 0.0.0.0, from 0.0.0.0, 00:00:00 ago, via GigabitEthernet0/0/0, prefer-non-rib-labels, merge-labels
        # 0.0.0.0, from 0.0.0.0, 00:00:00 ago, via GigabitEthernet0/0/0
        p5 = re.compile(r'^\*? *(?P<nexthop>[\w\.]+)(?:, +from +(?P<from>[\w\.]+), +'
                         '(?P<age>[\w\.\:]+) +ago, +via +(?P<interface>\S+)(?:, +'
                         '(?P<rib_labels>prefer-non-rib-labels))?(?:, +'
                         '(?P<merge_labels>merge-labels))?)?$')
        p6 = re.compile(r'^Route +metric +is +(?P<metric>\d+), +'
                         'traffic +share +count +is +(?P<share_count>\d+)$')

        # ipv6 specific
        p7 = re.compile(r'^Route +count +is +(?P<route_count>[\d\/]+), +'
                         'share +count +(?P<share_count>[\d\/]+)$')
        p8 = re.compile(r'^(?P<fwd_ip>[\w\:]+)(, +(?P<fwd_intf>[\w\.\/\-]+)'
                         '( indirectly connected)?)?$')
        p8_1 = re.compile(r'^receive +via +(?P<fwd_intf>[\w\.\/\-]+)$')
        p9 = re.compile(r'^Last +updated +(?P<age>[\w\:\.]+) +ago$')
        p10 = re.compile(r'^From +(?P<from>[\w\:]+)$')

        # MPLS label: implicit-null
        p11 = re.compile(r'^MPLS +label: +(?P<mpls_label>\S+)$')

        # MPLS Flags: NSF
        p12 = re.compile(r'^MPLS +Flags: +(?P<mpls_flags>\S+)$')

        # SR Incoming Label: 00000
        p13 = re.compile(r'^SR +Incoming +Label: +(?P<sr_incoming_label>\d+)')

        # Repair Path: 0.0.0.0, via GigabitEthernet0
        p14 = re.compile(r'^Repair +Path: +(?P<path>[\d\.]+), +via +(?P<via>\w+)')

        # initial variables
        ret_dict = {}
        index = 0

        for line in out.splitlines():
            line = line.strip()

            # Routing entry for 10.151.0.0/24, 1 known subnets
            # Routing entry for 0.0.0.0/0, supernet
            # Routing entry for 192.168.154.0/24
            m = p1.match(line)
            if m:
                group = m.groupdict()
                entry = group.pop('entry')
                entry_dict = ret_dict.setdefault('entry', {}).setdefault(entry, {})
                entry_dict.update({k:v for k,v in group.items() if v})
                continue

            # Known via "static", distance 1, metric 0, candidate default path
            # Known via "eigrp 1", distance 130, metric 10880, type internal
            # Known via "rip", distance 120, metric 2
            # Known via "connected", distance 0, metric 0 (connected)
            # Known via "eigrp 1", distance 130, metric 10880, type internal
            m = p2.match(line)
            if m:
                group = m.groupdict()
                entry_dict.update({k:v for k,v in group.items() if v})
                continue

            # Redistributing via rip
            # Redistributing via eigrp 1
            m = p3.match(line)
            if m:
                group = m.groupdict()
                entry_dict.update({k:v for k,v in group.items() if v})
                continue

            # Last update from 192.168.151.2 on Vlan101, 2w3d ago
            # Last update from 192.168.246.2 on Vlan103, 00:00:12 ago
            m = p4.match(line)
            if m:
                group = m.groupdict()
                update_dict = entry_dict.setdefault('update', {})
                update_dict.update({k:v for k,v in group.items() if v})
                continue

            # * 192.168.151.2, from 192.168.151.2, 2w3d ago, via Vlan101
            # * 10.69.1.2
            # 0.0.0.0, from 0.0.0.0, 00:00:00 ago, via GigabitEthernet0/0/0, prefer-non-rib-labels, merge-labels
            # 0.0.0.0, from 0.0.0.0, 00:00:00 ago, via GigabitEthernet0/0/0
            m = p5.match(line)
            if m:
                group = m.groupdict()
                index += 1
                path_dict = entry_dict.setdefault('paths', {}).setdefault(index, {})

                if group['nexthop']:
                    path_dict.update({'nexthop': group['nexthop']})
                if group['from']:
                    path_dict.update({'from': group['from']})
                if group['age']:
                    path_dict.update({'age': group['age']})
                if group['interface']:
                    path_dict.update({'interface': group['interface']})

                path_dict.update({'prefer_non_rib_labels': True if group['rib_labels'] else False})
                path_dict.update({'merge_labels': True if group['merge_labels'] else False})
                continue

            # Route metric is 10880, traffic share count is 1
            m = p6.match(line)
            if m:
                group = m.groupdict()
                path_dict = entry_dict.setdefault('paths', {}).setdefault(index, {})
                path_dict.update({k:v for k,v in group.items() if v})
                continue

            # Route count is 1/1, share count 0
            m = p7.match(line)
            if m:
                group = m.groupdict()
                entry_dict.update({k:v for k,v in group.items() if v})
                continue

            # FE80::EEBD:1DFF:FE09:56C2, Vlan202
            # FE80::EEBD:1DFF:FE09:56C2
            m = p8.match(line)
            if m:
                group = m.groupdict()
                index += 1
                path_dict = entry_dict.setdefault('paths', {}).setdefault(index, {})
                path_dict.update({k:v for k,v in group.items() if v})
                continue

            # receive via Loopback4
            m = p8_1.match(line)
            if m:
                group = m.groupdict()
                index += 1
                path_dict = entry_dict.setdefault('paths', {}).setdefault(index, {})
                path_dict.update({k:v for k,v in group.items() if v})
                continue

            # From FE80::EEBD:1DFF:FE09:56C2
            m = p10.match(line)
            if m:
                path_dict['from'] = m.groupdict()['from']
                continue

            # Last updated 2w4d ago
            m = p9.match(line)
            if m:
                path_dict['age'] = m.groupdict()['age']
                continue

            # MPLS label: implicit-null
            m = p11.match(line)
            if m:
                path_dict = entry_dict.setdefault('paths', {}).setdefault(index, {})
                path_dict.update({'mpls_label': m.groupdict()['mpls_label']})
                continue

            # MPLS Flags: NSF
            m = p12.match(line)
            if m:
                path_dict = entry_dict.setdefault('paths', {}).setdefault(index, {})
                path_dict.update({'mpls_flags': m.groupdict()['mpls_flags']})
                continue

            # SR Incoming Label: 00000
            m = p13.match(line)
            if m:
                entry_dict.update({'sr_incoming_label': m.groupdict()['sr_incoming_label']})
                continue

            # Repair Path: 0.0.0.0, via GigabitEthernet0
            m = p14.match(line)
            if m:
                path_dict = entry_dict.setdefault('paths', {}).setdefault(index, {}).setdefault('repair_path', {})
                path_dict.update({'repair_path': m.groupdict()['path']})
                path_dict.update({'via': m.groupdict()['via']})


        ret_dict.update({'total_prefixes': index}) if ret_dict else None
        return ret_dict


# ====================================================
#  schema for show ipv6 route <WORD>
# ====================================================
class ShowIpv6RouteWordSchema(MetaParser):
    """Schema for show ipv6 route <WORD>"""
    schema = {
        'entry': {
            Any(): {
                'ip': str,                
                'mask': str,
                'known_via': str,
                'distance': str,
                'metric': str,
                Optional('route_count'): str,
                Optional('share_count'): str,
                Optional('type'): str,
                'paths': {
                    Any(): {
                        Optional('fwd_ip'): str,
                        Optional('fwd_intf'): str,
                        Optional('from'): str,
                        Optional('age'): str
                    }
                }
            }
        },
        'total_prefixes': int,
    }


# ====================================================
#  parser for show ipv6 route <WORD>
# ====================================================
class ShowIpv6RouteWord(ShowIpv6RouteWordSchema, ShowIpRouteWord):
    """Parser for :
       show ipv6 route <Hostname or A.B.C.D>
       show ipv6 route vrf <vrf> <Hostname or A.B.C.D>"""
    command = ['show ipv6 route vrf {vrf}', 'show ipv6 route vrf {vrf} {route}',
                   'show ipv6 route', 'show ipv6 route {route}']
    IP_VER = 'ipv6'

    def cli(self, route='', vrf='', output=None):
        if not vrf:
            vrf = 'default'
        if output is None:
            if vrf != 'default':
                if route:
                    cmd = self.command[1].format(vrf=vrf, route=route)
                else:
                    cmd = self.command[0].format(vrf=vrf)
            else:
                if route:
                    cmd = self.command[3].format(route=route)
                else:
                    cmd = self.command[2]
            out = self.device.execute(cmd)
        else:
            out = output

        return super().cli(route=route, vrf=vrf, output=out)


# ====================================================
#  schema for show ip cef
# ====================================================
class ShowIpCefSchema(MetaParser):
    """Schema for show ip cef show ip cef
                  show ip cef vrf <vrf>
                  show ip cef <prefix>
                  show ip cef <prefix> detail
                  show ip cef vrf <vrf> <prefix>"""
    schema = {
        'vrf':
            {Any():
                {'address_family':
                    {Any():
                        {'prefix':
                            {Any():
                                {'nexthop':
                                    {Any():
                                        {Optional('outgoing_interface'):
                                            {Any():
                                                {Optional('local_label'): int,
                                                Optional('outgoing_label'): list,
                                                Optional('repair'): str,
                                                },
                                            },
                                        },
                                    },
                                Optional('epoch'): int,
                                Optional('per_destination_sharing'): bool,
                                Optional('sr_local_label_info'): str,
                                },
                            },
                        },
                    },
                },
            },
        }

# ====================================================
#  parser  for show ip cef <ip>
# ====================================================
class ShowIpCef(ShowIpCefSchema):
    """parser for show ip cef
                  show ip cef vrf <vrf>
                  show ip cef <prefix>
                  show ip cef <prefix> detail
                  show ip cef vrf <vrf> <prefix>"""

    cli_command = ['show ip cef',
                   'show ip cef vrf {vrf}',
                   'show ip cef {prefix}',
                   'show ip cef vrf {vrf} {prefix}']

    def cli(self, vrf="", prefix="", cmd="", output=None):

        if output is None:
            if not cmd:
                if vrf:
                    if prefix:
                        cmd = self.cli_command[3].format(vrf=vrf,prefix=prefix)
                    else:
                        cmd = self.cli_command[1].format(vrf=vrf)
                else:
                    if prefix:
                        cmd = self.cli_command[2].format(prefix=prefix)
                    else:
                        cmd = self.cli_command[0]

            out = self.device.execute(cmd)
        else:
            out = output

        if not vrf:
            vrf = 'default'

        # initial return dictionary
        result_dict = {}

        # 10.169.197.104/30
        # 2001:DB8:1:3::/64
        # 2.2.2.2/32, epoch 2, per-destination sharing
        p1 = re.compile(r'^(?P<prefix>[\w\:\.]+[\/]+[\d]+)'
                         '(?:, +epoch +(?P<epoch>(\d+)))?'
                         '(?:, +(?P<sharing>(per-destination sharing)))?$')

        # sr local label info: global/16002 [0x1B]
        p1_1 = re.compile(r'^sr +local +label +info: +(?P<sr_local_label_info>(.*))$')

        #     nexthop 10.169.197.93 TenGigabitEthernet0/2/0 label 22-(local:2043)
        #     nexthop 10.1.2.2 GigabitEthernet2.100
        #     nexthop FE80::A8BB:CCFF:FE03:2101 FastEthernet0/0/0 label 18
        #     nexthop 10.2.3.3 FastEthernet1/0/0 label 17 24
        p2 = re.compile(r'^nexthop +(?P<nexthop>[\w\.\:]+) +(?P<interface>\S+)'
                       '( +label +(?P<outgoing_label>[\w\-\ ]+)(-\(local:(?P<local_label>\w+)\))?)?$')

        #     nexthop 10.0.0.5 GigabitEthernet2 label [16002|16002]-(local:16002)
        p2_1 = re.compile(r'^nexthop +(?P<nexthop>\S+) +(?P<interface>\S+)'
                           ' +label +\[(?P<outgoing_label>([0-9\|]+))\]'
                           '\-\(local\:(?P<local_label>(\d+))\)$')

        #     attached to GigabitEthernet3.100
        p3 = re.compile(r'^(?P<nexthop>\w+) +(to|for) +(?P<interface>\S+)$')

        #  no route
        p4 = re.compile(r'^(?P<nexthop>[a-z\ ]+)$')

        # 10.1.2.255/32        receive              GigabitEthernet2.100
        # 10.1.3.0/24          10.1.2.1             GigabitEthernet2.100
        #                      10.2.3.3             GigabitEthernet3.100
        p5 = re.compile(r'^((?P<prefix>[\w\:\.]+[\/]+[\d]+) +)?(?P<nexthop>[\w\.]+)( +(?P<interface>[^a-z][\S]+))?$')

        # repair: attached-nexthop 10.0.0.9 GigabitEthernet3
        p6 = re.compile(r'^repair: +(?P<repair>(.*))$')

        for line in out.splitlines():
            line = line.strip()

            # 10.169.197.104/30
            # 2001:DB8:1:3::/64
            # 2.2.2.2/32, epoch 2, per-destination sharing
            m = p1.match(line)
            if m:
                group = m.groupdict()
                if ':' in group['prefix']:
                    address_family = 'ipv6'
                else:
                    address_family = 'ipv4'
                prefix_dict = result_dict.setdefault('vrf',{}).\
                                          setdefault(vrf, {}).\
                                          setdefault('address_family', {}).\
                                          setdefault(address_family,{}).\
                                          setdefault('prefix',{}).\
                                          setdefault(group['prefix'], {})
                if group['epoch']:
                    prefix_dict['epoch'] = int(group['epoch'])
                if group['sharing']:
                    prefix_dict['per_destination_sharing'] = True
                continue

            # sr local label info: global/16002 [0x1B]
            m = p1_1.match(line)
            if m:
                prefix_dict['sr_local_label_info'] = m.groupdict()['sr_local_label_info']
                continue

            #   nexthop 10.169.197.93 TenGigabitEthernet0/2/0 label 22-(local:2043)
            #   nexthop 10.1.2.2 GigabitEthernet2.100
            m = p2.match(line)
            if m:
                group = m.groupdict()
                nexthop_dict = prefix_dict.setdefault('nexthop', {}).\
                                           setdefault(group['nexthop'], {}).\
                                           setdefault('outgoing_interface', {}).\
                                           setdefault(group['interface'], {})

                if group['local_label']:
                    nexthop_dict.update({'local_label': int(group['local_label'])})
                if group['outgoing_label']:
                    nexthop_dict.update({'outgoing_label': group['outgoing_label'].split()})
                continue

            #     nexthop 10.0.0.5 GigabitEthernet2 label [16002|16002]-(local:16002)
            m = p2_1.match(line)
            if m:
                group = m.groupdict()
                nexthop_dict = prefix_dict.setdefault('nexthop', {}).\
                                           setdefault(group['nexthop'], {}).\
                                           setdefault('outgoing_interface', {}).\
                                           setdefault(group['interface'], {})

                nexthop_dict.update({'local_label': int(group['local_label'])})
                nexthop_dict.update({'outgoing_label': group['outgoing_label'].split()})
                continue

            # attached to GigabitEthernet3.100
            m = p3.match(line)
            if m:
                group = m.groupdict()
                prefix_dict.setdefault('nexthop', {}). \
                            setdefault(group['nexthop'], {}). \
                            setdefault('outgoing_interface', {}). \
                            setdefault(group['interface'], {})
                continue

            #  no route
            #  discard
            m = p4.match(line)
            if m:
                group = m.groupdict()
                prefix_dict.setdefault('nexthop', {}). \
                            setdefault(group['nexthop'], {})

                continue

            # 10.1.2.255/32        receive              GigabitEthernet2.100
            # 10.1.3.0/24          10.1.2.1             GigabitEthernet2.100
            #                      10.2.3.3             GigabitEthernet3.100
            m = p5.match(line)
            if m:
                group = m.groupdict()
                if group['prefix']:
                    prefix = group['prefix']

                    if ':' in group['prefix']:
                        address_family = 'ipv6'
                    else:
                        address_family = 'ipv4'
                prefix_dict = result_dict.setdefault('vrf', {}). \
                    setdefault(vrf, {}). \
                    setdefault('address_family', {}). \
                    setdefault(address_family, {}). \
                    setdefault('prefix', {}). \
                    setdefault(prefix, {}).\
                    setdefault('nexthop', {}).\
                    setdefault(group['nexthop'], {})
                if group['interface']:
                    prefix_dict.setdefault('outgoing_interface', {}).\
                                setdefault(group['interface'], {})
                continue

            # repair: attached-nexthop 10.0.0.9 GigabitEthernet3
            m = p6.match(line)
            if m:
                nexthop_dict.update({'repair': m.groupdict()['repair']})
                continue

        return result_dict


# ====================================================
#  parser  for show ipv6 cef
# ====================================================
class ShowIpv6Cef(ShowIpCef):
    """parser for show ipv6 cef
                  show ipv6 cef vrf <vrf>
                  show ipv6 cef <prefix>
                  show ipv6 cef vrf <vrf> <prefix>"""

    cli_command = ['show ipv6 cef',
                   'show ipv6 cef vrf {vrf}',
                   'show ipv6 cef {prefix}',
                   'show ipv6 cef vrf {vrf} {prefix}']

    def cli(self, vrf="", prefix="", cmd="", output=None):

        if output is None:
            if vrf:
                if prefix:
                    cmd = self.cli_command[3].format(vrf=vrf,prefix=prefix)
                else:
                    cmd = self.cli_command[1].format(vrf=vrf)
            else:
                if prefix:
                    cmd = self.cli_command[2].format(prefix=prefix)
                else:
                    cmd = self.cli_command[0]
        else:
            output = output

        return super().cli(cmd=cmd, vrf=vrf, prefix=prefix, output=output)


# =========================================
#  Parser for 'show ip cef <prefix> detail'
# =========================================
class ShowIpCefDetail(ShowIpCef):
    ''' Parser for:
        * 'show ip cef <prefix> detail'
    '''

    cli_command = 'show ip cef {prefix} detail'

    def cli(self, prefix, output=None):

        if output is None:
            output = self.device.execute(self.cli_command.format(prefix=prefix))
        else:
            output = output

        return super().cli(prefix=prefix, output=output)


# ====================================================
#  schema for show ip route summary
# ====================================================
class ShowIpRouteSummarySchema(MetaParser):
    """Schema for show ip route summary
                  show ip route vrf <vrf> summary
    """
    schema = {
        'vrf':{
        Any():{
            'vrf_id': str,
            'maximum_paths': int,
            'total_route_source': {
                'networks': int,
                'subnets': int,
                'replicates': int,
                'overhead': int,
                'memory_bytes': int,
            },
            'route_source': {
                Any(): {

                        Any(): {
                            'networks': int,
                            'subnets': int,
                            'replicates': int,
                            'overhead': int,
                            'memory_bytes': int,
                            Optional('intra_area'): int,
                            Optional('inter_area'): int,
                            Optional('external_1'): int,
                            Optional('external_2'): int,
                            Optional('nssa_external_1'): int,
                            Optional('nssa_external_2'): int,
                            Optional('level_1'): int,
                            Optional('level_2'): int,
                            Optional('external'): int,
                            Optional('internal'): int,
                            Optional('local'): int,

                    },
                    Optional('networks'): int,
                    Optional('subnets'): int,
                    Optional('replicates'): int,
                    Optional('overhead'): int,
                    Optional('memory_bytes'): int,
                    Optional('intra_area'): int,
                    Optional('inter_area'): int,
                    Optional('external_1'): int,
                    Optional('external_2'): int,
                    Optional('nssa_external_1'): int,
                    Optional('nssa_external_2'): int,
                    Optional('level_1'): int,
                    Optional('level_2'): int,
                    Optional('external'): int,
                    Optional('internal'): int,
                    Optional('local'): int,
                },

            }

        }
    }
    }

# ====================================================
#  parser for show ip route summary
# ====================================================
"""Parser for show ip route summary
                  show ip route vrf <vrf> summary
    """
class ShowIpRouteSummary(ShowIpRouteSummarySchema):

    cli_command = ['show ip route summary', 'show ip route vrf {vrf} summary']

    def cli(self, vrf='', output=None):
        if output is None:
            if vrf:
                cmd = self.cli_command[1].format(vrf=vrf)
            else:
                cmd = self.cli_command[0]
            out = self.device.execute(cmd)
        else:
            out = output

        # IP routing table name is default (0x0)
        # IP routing table name is VRF-1 (0x27)
        p1 = re.compile(r'^IP +routing +table +name +is +(?P<vrf>\S+) +\((?P<vrf_id>\w+)\)$')
        # IP routing table maximum-paths is 32
        p2 = re.compile(r'^IP +routing +table +maximum-paths +is +(?P<max_path>[\d]+)$')
        # application     0           0           0           0           0
        p3 = re.compile(
            r'^(?P<protocol>\w+) +(?P<instance>\w+)*? *(?P<networks>\d+) +('
            r'?P<subnets>\d+)? +(?P<replicates>\d+)? +(?P<overhead>\d+)? +('
            r'?P<memory_bytes>\d+)$')
        # Intra-area: 1 Inter-area: 0 External-1: 0 External-2: 0
        p7 = re.compile(
            r'^Intra-area: +(?P<intra_area>\d+) +Inter-area: +(?P<inter_area>\d+) '
            r'+External-1: +(?P<external_1>\d+) +External-2: +(?P<external_2>\d+)$')
        #   NSSA External-1: 0 NSSA External-2: 0
        p8 = re.compile(
            r'^NSSA +External-1: +(?P<nssa_external_1>\d+) +NSSA +External-2: +('
            r'?P<nssa_external_2>\d+)$')
        #   Level 1: 1 Level 2: 0 Inter-area: 0
        p9_1 = re.compile(
            r'^Level +1: +(?P<level_1>\d+) +Level +2: +(?P<level_2>\d+) +Inter-area: +('
            r'?P<inter_area>\d+)$')
        #   External: 0 Internal: 0 Local: 0
        p13 = re.compile(
            r'^External: +(?P<external>\d+) +Internal: +(?P<internal>\d+) +Local: +(?P<local>\d+)$')

        ret_dict = {}
        for line in out.splitlines():
            line = line.strip()
            # IP routing table name is default (0x0)
            m = p1.match(line)
            if m:
                vrf = m.groupdict()['vrf']
                vrf_dict = ret_dict.setdefault('vrf',{}).setdefault(vrf, {})
                vrf_dict['vrf_id'] = m.groupdict()['vrf_id']
                vrf_rs_dict = vrf_dict.setdefault('route_source', {})
                continue
            # IP routing table maximum-paths is 32
            m = p2.match(line)
            if m:
                vrf_dict['maximum_paths'] = int(m.groupdict()['max_path'])
                continue
            # application     0           0           0           0           0
            m = p3.match(line)
            if m:
                group = m.groupdict()
                protocol = group.pop('protocol')
                instance = group.pop('instance')
                if protocol == 'Total':
                    protocol_dict = vrf_dict.setdefault('total_route_source', {})
                else:
                    protocol_dict = vrf_rs_dict.setdefault(protocol, {})
                if instance is not None:
                    inst_dict = protocol_dict.setdefault(instance, {})
                    inst_dict.update({k:int(v) for k, v in group.items() if v is not None})
                else:
                    group = {k: int(v) for k, v in group.items() if v is not None}
                    protocol_dict.update(group)
                continue
            # Intra-area: 1 Inter-area: 0 External-1: 0 External-2: 0
            m = p7.match(line)
            if m:
                group = {k: int(v) for k, v in m.groupdict().items()}
                vrf_rs_dict.setdefault('ospf', {})
                vrf_rs_dict['ospf'][instance].update(group)
                continue
            #   NSSA External-1: 0 NSSA External-2: 0
            m = p8.match(line)
            if m:
                group = {k: int(v) for k, v in m.groupdict().items()}
                vrf_rs_dict.setdefault('ospf', {})
                vrf_rs_dict['ospf'][instance].update(group)
                continue

            #   Level 1: 1 Level 2: 0 Inter-area: 0
            m = p9_1.match(line)
            if m:
                group = {k: int(v) for k, v in m.groupdict().items()}
                vrf_rs_dict.setdefault('isis', {})
                vrf_rs_dict['isis'][instance].update(group)
                continue

            #   External: 0 Internal: 0 Local: 0
            m = p13.match(line)
            if m:
                group = {k: int(v) for k, v in m.groupdict().items()}
                vrf_rs_dict.setdefault('bgp', {})
                vrf_rs_dict['bgp'][instance].update(group)
                continue
        return ret_dict
